;    This file is part of CrabEmu BIOS.
;
;    Copyright (C) 2008, 2009 Lawrence Sebald
;
;    CrabEmu BIOS is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License version 2 
;    as published by the Free Software Foundation.
;
;    CrabEmu BIOS is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with CrabEmu BIOS; if not, write to the Free Software
;    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

; Some constants to start with.
    .define     ROM_Addr    $0000
    .define     RAM_Addr    $C000
    .define     RAM_Len     $2000
    .define     Memctl      $3E
    .define     Memctl_None $EB
    .define     Memctl_BIOS $E3
    .define     Memctl_Card $CB
    .define     Memctl_Cart $AB
    .define     Memctl_Exp  $6B

; These are placed here by the startup code.
    .enum $C000
    Last_Memctl     db
    Checksum_Return db
    Control_Data    db
    Region_Code     db
    .ende

; Set up the memory map needed by WLA DX.
    .memorymap
    defaultslot     0
    slotsize        $1000
    slot            0           RAM_Addr + (RAM_Len / 2)
    .endme

    .rombankmap
    bankstotal      1
    banksize        $1000
    banks           1
    .endro

; The start of the export checksum handling code. This is run from address D000.
    .bank 0 slot 0
    .orga   RAM_Addr + (RAM_Len / 2)
start_checksum:
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Card      ; Try the card slot first.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    find_tmr_sega
    jr      z, +                ; If we've found a header, continue onward.
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Cart      ; Otherwise, try the cartridge slot next.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    find_tmr_sega
    jr      z, +                ; If we've found a header, continue onward.
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Exp       ; Otherwise, try the expansion slot last.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    find_tmr_sega
    jr      z, +                ; If we've found a header, continue onward.
    ld      a, Memctl_BIOS
    out     (Memctl), a
    ld      a, Memctl_Cart      ; XXXX: Set this to cartridge...
    ld      (Last_Memctl), a
    ret                         ; We have no headers, bail.
+:
    push    bc
    push    de
    push    hl
    ld      bc, $f
    add     hl, bc
    ld      a, (hl)             ; Grab the last byte of the header
    and     $f                  ; We only want the low nibble
    ld      b, a                ; Multiply by 6
    rlc     a
    add     a, b
    rlc     a
    ld      hl, checksum_lut
    ld      c, a
    ld      b, 0
    add     hl, bc
    ld      c, (hl)             ; Grab the lower range for checksumming
    inc     hl
    ld      b, (hl)
    inc     hl
    push    hl
    ld      hl, 0
    ld      de, 0               ; Reset de to 0 before checksumming.
    call    checksum_range
    pop     hl                  ; Grab the upper range
    ld      c, (hl)
    inc     hl
    ld      b, (hl)
    inc     hl
    ld      a, (hl)
    or      b
    or      c
    jr      z, ++
    ld      (checksum_left), bc ; Save what we need to checksum still
    ld      a, (hl)
    ld      (checksum_left2), a
    ld      hl, (checksum_left)
-:
    scf
    ccf
    ld      bc, $4000
    sbc     hl, bc
    ld      (checksum_left), hl ; Save what we'll still need to checksum
    sbc     a, 0
    ld      (checksum_left2), a
    ld      hl, $8000
    ld      a, (checksum_page)  ; Grab the page we're checksumming
    ld      ($FFFF), a          ; And set the paging register
    inc     a
    ld      (checksum_page), a  ; Save the value for the next loop.
    call    checksum_range      ; Call the checksum routine
    ld      a, (checksum_left2)
    ld      hl, (checksum_left)
    or      h                   ; See if we have anything left
    or      l
    ld      a, (checksum_left2)
    jr      nz, -               ; Loop if we need to.
++:
    pop     hl
    ld      c, 10
    ld      b, 0
    add     hl, bc
    ld      a, (hl)             ; Compare calculated to the header's value.
    cpl
    and     e
    jr      nz, +++
    inc     hl
    ld      a, (hl)
    cpl
    and     d
    jr      nz, +++
    ld      a, Memctl_BIOS      ; If we get here, we have a valid checksum.
    out     (Memctl), a
    ld      a, 0
    ld      (Checksum_Return), a
    pop     de
    pop     bc
    ret
+++:
    ld      a, Memctl_BIOS
    out     (Memctl), a
    ld      a, -2
    ld      (Checksum_Return), a
    pop     de
    pop     bc
    ret

; If Z is set when this function returns, the "TMR SEGA" string was found
; at the address of hl.
check_hl:
    push    hl
    push    bc
    push    ix
    ld      bc, 8
    ld      ix, _tmr_sega_const
    ld      a, (ix + 0)         ; Grab first character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 1)         ; Grab second character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 2)         ; Grab third character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 3)         ; Grab fourth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 4)         ; Grab fifth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 5)         ; Grab sixth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 6)         ; Grab seventh character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 7)         ; Grab eighth character from reference
    cpi
+:
    pop     ix
    pop     bc
    pop     hl
    ret

; Attempt to find the TMR SEGA string that starts the standard SMS header.
find_tmr_sega:
    ld      hl, $1FF0           ; Check at 0x1FF0 first.
    call    check_hl
    jr      z, +
    ld      hl, $3FF0           ; Then try 0x3FF0.
    call    check_hl
    jr      z, +
    ld      hl, $7FF0           ; Finally, try 0x7FF0.
    call    check_hl
    jr      z, +
    ld      a, -1               ; We haven't found a header if we get here.
    ld      (Checksum_Return), a; Set a failure flag at 0xC001
    ret
+:
    xor     a                   ; We have found a header if we get here.
    ld      (Checksum_Return), a; Set a success flag at 0xC001
    ret

; Compute the checksum over a given range of the ROM.
checksum_range:                 ; Start addr in hl, size in bc, seed in de
    ld      a, e
    add     a, (hl)             ; Add the byte at hl to e, preserving carry
    ld      e, a
    ld      a, d
    adc     a, 0                ; Add the carry to d
    ld      d, a
    dec     bc
    ld      a, b                ; Check if bc == 0
    or      c
    inc     hl                  ; Increment the pointer.
    ret     z                   ; Return if bc == 0
    jr      checksum_range      ; Loop if bc != 0

; Lookup table for checksum ranges.
checksum_lut:
    .dw     $7FF0, $8000, $0003 ; Value 0 (256KB)
    .dw     $7FF0, $8000, $0007 ; Value 1 (512KB)
    .dw     $7FF0, $8000, $000F ; Value 2 (1024KB)
    .dw     $0000, $0000, $0000 ; Value 3 (invalid)
    .dw     $0000, $0000, $0000 ; Value 4 (invalid)
    .dw     $0000, $0000, $0000 ; Value 5 (invalid)
    .dw     $0000, $0000, $0000 ; Value 6 (invalid)
    .dw     $0000, $0000, $0000 ; Value 7 (invalid)
    .dw     $0000, $0000, $0000 ; Value 8 (invalid)
    .dw     $0000, $0000, $0000 ; Value 9 (invalid)
    .dw     $1FF0, $0000, $0000 ; Value 10 (8KB)
    .dw     $3FF0, $0000, $0000 ; Value 11 (16KB)
    .dw     $7FF0, $0000, $0000 ; Value 12 (32KB)
    .dw     $BFF0, $0000, $0000 ; Value 13 (48KB)
    .dw     $7FF0, $8000, $0000 ; Value 14 (64KB)
    .dw     $7FF0, $8000, $0001 ; Value 15 (128KB)

_tmr_sega_const:
    .db    "TMR SEGA"

; Temporary storage for doing the checksum calculation. Remember, this is in
; RAM when it runs.
checksum_left:
    .dw     0
checksum_left2:
    .db     0
checksum_page:
    .db     2
