;    This file is part of CrabEmu BIOS.
;
;    Copyright (C) 2008, 2009 Lawrence Sebald
;
;    CrabEmu BIOS is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License version 2 
;    as published by the Free Software Foundation.
;
;    CrabEmu BIOS is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with CrabEmu BIOS; if not, write to the Free Software
;    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

; Some constants to start with.
    .define     ROM_Addr    $0000
    .define     RAM_Addr    $C000
    .define     RAM_Len     $2000
    .define     Memctl      $3E
    .define     Memctl_None $EB
    .define     Memctl_BIOS $E3
    .define     Memctl_Card $CB
    .define     Memctl_Cart $AB
    .define     Memctl_Exp  $6B

; These are placed here by the startup code.
    .enum $C000
    Last_Memctl     db
    Checksum_Return db
    Control_Data    db
    Region_Code     db
    .ende

; Set up the memory map needed by WLA DX.
    .memorymap
    defaultslot     0
    slotsize        $1000
    slot            0           RAM_Addr + (RAM_Len / 2)
    .endme

    .rombankmap
    bankstotal      1
    banksize        $1000
    banks           1
    .endro

; The start of the "domestic" slot checking code. This is run from address D000.
    .bank 0 slot 0
    .orga   RAM_Addr + (RAM_Len / 2)
start_check:
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Card      ; Try the card slot first.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    check_slot
    jr      nz, +               ; If we've found data, continue onward.
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Cart      ; Otherwise, try the cartridge slot next.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    check_slot
    jr      nz, +               ; If we've found data, continue onward.
    ld      a, Memctl_None
    out     (Memctl), a
    ld      a, Memctl_Exp       ; Otherwise, try the expansion slot last.
    out     (Memctl), a
    ld      (Last_Memctl), a
    call    check_slot
    jr      nz, +               ; If we've found data, continue onward.
    ld      a, Memctl_BIOS
    out     (Memctl), a
    ld      a, Memctl_Cart      ; XXXX: Set this to cartridge...
    ld      (Last_Memctl), a
    ld      a, -3
    ld      (Checksum_Return), a
    ret                         ; We have no data, bail.
+:
    ld      a, Memctl_BIOS
    out     (Memctl), a
    xor     a
    ld      (Checksum_Return), a
    ret

; This code is adapted from code found on the following page:
; http://www.smspower.org/dev/docs/wiki/Software/BIOS
; under the heading "Japanese Master System BIOS v2.1 (1987)"
check_slot:
    push    hl                  ; Push registers on the stack that we'll use.
    push    bc
    push    de
    ld      hl, $0000           ; Start at address 0 of rom.
    ld      bc, $FF3C           ; Loop 255 times, set the mask to 3C.
    ld      a, (hl)             ; Grab the byte.
    and     c                   ; AND with 3C.
    ld      e, a                ; Store the result in e.
    inc     l                   ; Increment the pointer by 1.
-:
    ld      a, (hl)             ; Grab the next byte.
    inc     l                   ; Increment the pointer.
    and     c                   ; AND with 3C.
    cp      e                   ; Compare to first byte in e.
    jr      nz, +               ; If different, we've found something.
    djnz    -                   ; Otherwise, try again.
+:
    pop     de                  ; Clean up the stack, and return.
    pop     bc
    pop     hl
    ret

; Meaningless comment added because WLA-DX was being a bit odd, and
; segfaulting with just the code above.
; Meaningless comment added because WLA-DX was being a bit odd, and
; segfaulting with just the code above.
; Meaningless comment added because WLA-DX was being a bit odd, and
; segfaulting with just the code above.
